# -*- coding: utf-8 -*-
import os
import stat
import shutil
import logging as log
from glob import glob
from cStringIO import StringIO
from os.path import join, isdir, isfile

__doc__ = """SSH key stuff

    >>> HGAPP_AUTHORIZED_KEYS = '/tmp/hgapp_keys'
    >>> os.environ['HGAPP_AUTHORIZED_KEYS'] = HGAPP_AUTHORIZED_KEYS

Create user dir::

    >>> user_dir('gawel')
    '/tmp/hgapp_keys/gawel'
    >>> isdir(join(HGAPP_AUTHORIZED_KEYS, 'gawel'))
    True

List users::
    >>> users()
    ['gawel']

Add a key to the user::

    >>> keys = add_key('gawel', 'test_key', StringIO('ssh-dss %s= gael@hostname' % ('AAA'*255,)))
    >>> isfile(join(HGAPP_AUTHORIZED_KEYS, 'gawel', 'test_key.pub'))
    True
    >>> print open(keys[0]).read() #doctest: +ELLIPSIS
    ssh-dss AAA...A=

Get user keys::

    >>> user_keys('gawel') #doctest: +ELLIPSIS
    {'test_key': 'ssh-dss AAAAAAAAAAAAA...AAAAAAAA='}

Add another key::

    >>> keys = add_key('gawel', None, StringIO('ssh-dss %s= gael@hostname2' % ('AAA'*255,)))

Gen key file::

    >>> filename = join(HGAPP_AUTHORIZED_KEYS, 'authorized_keys')
    >>> authorized_keys(filename)
    >>> print open(filename).read() #doctest: +ELLIPSIS
    command="/...bin/hgapp-ssh -u gawel",no-port-forwarding,no-agent-forwarding,no-X11-forwarding ssh-dss AAAA...AAAAAA= gael@hostname2
    command="/...bin/hgapp-ssh -u gawel",no-port-forwarding,no-agent-forwarding,no-X11-forwarding ssh-dss AAAA...AAAAAA= test_key
    <BLANKLINE>
    

Del a key::

    >>> del_key('gawel', 'test_key')
    >>> isfile(join(HGAPP_AUTHORIZED_KEYS, 'gawel', 'test_key.pub'))
    False

Del a user::

    >>> del_user('gawel')
    >>> isdir(join(HGAPP_AUTHORIZED_KEYS, 'gawel'))
    False

...
    >>> shutil.rmtree(HGAPP_AUTHORIZED_KEYS)

"""

SSH_KEY_TEMPLATE = '''command="%(binary)s -u %(user)s",''' + \
'''no-port-forwarding,no-agent-forwarding,no-X11-forwarding %(key)s %(name)s
'''

def chmod(*args):
    path = join(*args)
    if isdir(path):
        os.chmod(path, stat.S_IRWXU)
    else:
        os.chmod(path, stat.S_IRWXU)

def users():
    dirname = os.environ.get('HGAPP_AUTHORIZED_KEYS', os.path.expanduser('~/.ssh/hgapp'))
    if not os.path.isdir(dirname):
        os.makedirs(dirname)
    chmod(dirname)
    return sorted(os.listdir(dirname))

def user_dir(user):
    dirname = os.environ.get('HGAPP_AUTHORIZED_KEYS', os.path.expanduser('~/.ssh/hgapp'))
    if not os.path.isdir(dirname):
        os.makedirs(dirname)
    chmod(dirname)
    userdir = join(dirname, user)
    if not isdir(userdir):
        os.makedirs(userdir)
    chmod(userdir)
    return userdir

def normalize_name(name):
    if not name:
        return name
    for c in '~_./\\':
        name = name.replace(c, '_')
        name = name.strip(c)
    name = name.replace('__', '_')
    name = name.strip('_')
    return name

def add_key(user, name=None, key=None):
    if not hasattr(key, 'read'):
        if isfile(key):
            key = open(key, 'rb')
        else:
            raise IOError('No such file %s' % key)
    names = []
    filenames = []
    name = normalize_name(name)
    keys = [k for k in key.readlines() if k.strip()]
    log.debug('%i key(s) found', len(keys))
    for data in keys:
        data = data.strip()
        if ' ' in data:
            splited = data.split(' ')
            data = ' '.join(splited[:2])
            if len(splited) == 3 and not name:
                name = splited[2]
                name = normalize_name(name)
        if not name:
            raise RuntimeError('No name found in key. Use -n')
        if name in names:
            for i in range(50):
                name_ = '%s_%i' % (name, i)
                if name_ not in names:
                    name = name_
                    break
        names.append(name)
        if '\n' not in data and data.startswith('ssh-'):
            filename = join(user_dir(user), '%s.pub' % name)
            fd = open(filename, 'wb')
            fd.write(data)
            fd.close()
            filenames.append(filename)
        name = None
    return filenames

def gen_keys(fd):
    from hgapp import utils
    binary = utils.get_binary('ssh')
    for u in users():
        kwargs = dict(binary=binary, user=u)
        keys = user_keys(u, real=True).items()
        for k, v in sorted(keys):
            kwargs['name'] = k.strip()
            kwargs['key'] = v.strip()
            fd.write(SSH_KEY_TEMPLATE % kwargs)
    return fd

def authorized_keys(filename=None):
    fd = StringIO()
    gen_keys(fd)
    fd.seek(0)
    filename = filename or os.path.expanduser('~/.ssh/authorized_keys')
    keys = open(filename, 'wb')
    keys.write(fd.read())
    keys.close()
    chmod(filename)

def del_key(user, name):
    filename = join(user_dir(user), '%s.pub' % name)
    if os.path.isfile(filename):
        os.remove(filename)

def del_user(user):
    userdir = user_dir(user)
    shutil.rmtree(userdir)

def user_keys(user, real=False):
    keys = {}
    userdir = user_dir(user)
    for filename in glob(join(userdir, '*.pub')):
        fd = open(filename, 'rb')
        data = fd.read()
        if not real:
            data = '%s...%s' % (data[:30], data[-50:])
        keys[os.path.basename(filename)[:-4]] = data
        fd.close()
    return keys

def get_user_keys(user):
    return glob.glob(join(user_dir(user), '*.pub'))

