# -*- coding: utf-8 -*-
import logging as log
import sys
import os

def main():
    from optparse import OptionParser

    parser = OptionParser()
    parser.usage = """
        hg clone . ssh://[user@]host/path/to/repository
        hg [pull|push|...] ssh://[user@]host/path/to/repository

        ssh [user@]host help
        ssh [user@]host keys [options]
        ssh [user@]host [rebuild|sdist] [-v] /path/to/repository

SSH Commands:
  keys                  see options
  rebuild               rebuild docs for the repository
  sdist                 build sdist package"""
    parser.add_option("-u", "--user", dest="user",
                      default=None,
                      help="user name (only valid with command line)")
    parser.add_option('-n', "--name", dest="name",
                      default=None,
                      help="name associate to ssh key")
    parser.add_option('-k', "--key", dest="key",
                      default=None,
                      help="path to an ssh pub key to add at NAME for USER")
    parser.add_option('-d', "--delete", dest="delete",
                      action="store_true", default=False,
                      help="delete key NAME for USER")
    parser.add_option('-l', "--list", dest="list",
                      action="store_true", default=False,
                      help="list ssh keys")
    parser.add_option("-v", "--verbose", dest="verbose",
                      action="count", default=0)

    options, args = parser.parse_args()


    cmd = os.getenv('SSH_ORIGINAL_COMMAND', '?')
    path_info = None

    if cmd.startswith('hg init '):
        path_info = cmd[8:].strip()
    elif cmd.startswith('hg -R ') and cmd.endswith(' serve --stdio'):
        path_info = cmd[6:-14].strip('/')
    elif '-u' in cmd:
        parser.error("Illegal command %r" % cmd)
    else:
        cmd_args = sys.argv[1:]
        if cmd != '?':
            cmd_args.extend(cmd.split())
        options, args = parser.parse_args(cmd_args)
        log.basicConfig(stream=sys.stderr,
                        format=options.verbose and '%(levelname)-5.5s %(message)s' or '%(message)s',
                        level=options.verbose and log.DEBUG or log.INFO)
        log.debug('options: %s, args: %r', options, args)
        if len(args) == 2:
            if args[0] in ('rebuild', 'sdist'):
                cmd = args[0]
                path_info = args[1].strip('/').strip()
        elif args == ['help']:
            parser.parse_args(['-h'])
        elif args == ['keys']:
            from hgapp.script import ssh_keys
            ssh_keys(cmd_args)
            sys.exit(0)

    log.basicConfig(stream=sys.stderr,
                    format=options.verbose and '%(levelname)-5.5s %(message)s' or '%(message)s',
                    level=options.verbose and log.DEBUG or log.INFO)

    if not path_info:
        log.debug(path_info)
        parser.error("Illegal command %r" % cmd)
        sys.exit(-1)


    from hgapp import utils


    from hgapp import config

    repo = c = None

    repo, ui, c = utils.get_repo(path_info, user=options.user)

    if not repo and cmd.startswith('hg init'):
        repo = os.path.join(c.hgwebdir, path_info.split('/')[-1])
    elif not repo:
        parser.error("No valid repository found for %r" % path_info)

    c.repo = os.path.realpath(repo)

    if cmd.startswith('hg init'):
        args = ['init', c.repo]
    else:
        args = ['-R', c.repo, 'serve', '--stdio']


    if not c.disable_ssh_perms.as_bool('false'):
        if not config.has_section('filter:auth'):
            error = ("There is no filter:auth section in your hgapp.ini. "
                     "Please add one or set disable_ssh_perms to true")
            parser.error(error)

        from webob import Request
        from paste.deploy import loadfilter

        req = Request({
             'PATH_INFO': '/%s' % path_info,
             'REMOTE_USER':options.user,
             'HTTP_USER_AGENT': 'SSH',
             'REQUEST_METHOD':'POST',
             'SERVER_NAME':'localhost',
             'SERVER_PORT':'22',
             'wsgi.url_scheme':'https',
             'wsgi.version': '1.0',
            })
        req.hgapp_ui = ui
        auth_filter = loadfilter(
                'config:%s#auth' % config.defaults()['__file__'])
        identity = {'repoze.who.userid': options.user}
        auth_filter = auth_filter(object())
        classification = auth_filter.classifier(req.environ)
        auth_filter.add_metadata(req.environ, classification, identity)

        try:
            utils.check_perm(req, 'hg pull' in cmd and 'pull' or 'push')
        except Exception, e:
            parser.error('%s for %s' % (e.status, options.user))
        else:
            log.info('Access to %s granted for %s', req.path_info, options.user)

    if cmd in ('rebuild', 'sdist'):
        from mercurial import hg
        from hgapp import hooks
        repo = hg.repository(ui, c.repo)
        log.debug('Repo: %s', c.repo)
        if cmd == 'rebuild':
            log.info('Rebuilding docs for %s...' % path_info)
            hooks.sphinx_update(ui, repo, run_buildout=True,
                                hooktype='changegroup',
                                admin=True)
        else:
            log.info('Building sdist for %s...' % path_info)
            hooks.sdist_update(ui, repo, run_buildout=True,
                                hooktype='changegroup',
                                admin=True)
    else:
        from mercurial.dispatch import _runcatch
        _runcatch(ui, args)

