# -*- coding: utf-8 -*-
from ConfigObject import ConfigObject
from mercurial import config
from mercurial import ui
from webob import exc
from hgapp import config as hgconfig
import subprocess
import logging
import glob
import sys
import os

log = logging.getLogger(__name__)

def check_perm(req, op):
    '''Check permission for operation based on request data (including
    authentication info). Return if op allowed, else raise an ErrorResponse
    exception.'''

    user = req.remote_user
    c = req.hgapp_ui
    credentials = req.environ.get('repoze.what.credentials', {})
    groups = credentials.get('groups', ())
    permissions = credentials.get('permissions', ())

    if user and permissions:
        log.debug('check_perm %(PATH_INFO)s: %(REMOTE_USER)s - %(repoze.what.credentials)s',
                  req.environ)
    elif user:
        log.debug('check_perm %(PATH_INFO)s: %(REMOTE_USER)s', req.environ)

    require_groups = c.configlist('web', 'require_groups')
    if require_groups:
        if not set(require_groups).intersection(set(groups)):
            log.debug('require_groups failed')
            resp = exc.HTTPUnauthorized()
            resp.status = '%s one of this groups %s is required' % (resp.code, ', '.join(require_groups))
            raise resp

    def append(option):
        if user:
            value = c.configlist('web', option)
            value.append(user)
            value = ' '.join([i for i in value if i and i not in '-'])
            log.debug('Setting value %s to %r', option, value)
            c.setconfig('web', option, value)

    if user and 'allow_read' in permissions:
        append('allow_read')
    elif user and 'allow_read' not in permissions:
        deny_read = c.configlist('web', 'deny_read')
        if deny_read and (not user or deny_read == ['*'] or user in deny_read):
            log.debug('deny_read success')
            resp = exc.HTTPUnauthorized()
            resp.status = '%s %s' % (resp.code, 'read not authorized')
            raise resp

        allow_read = c.configlist('web', 'allow_read')
        result = (not allow_read) or (allow_read == ['*'])
        if not (result or user in allow_read):
            log.debug('allow_read failed')
            resp = exc.HTTPUnauthorized()
            resp.status = '%s %s' % (resp.code, 'read not authorized for %s' % user)
            raise resp
    else:
        allow_read = c.configlist('web', 'allow_read')
        if allow_read != ['*'] and user not in allow_read:
            log.debug('allow_read failed')
            resp = exc.HTTPUnauthorized()
            resp.status = '%s %s' % (resp.code, 'read not authorized for %s' % user)
            raise resp

    if op == 'pull' and not c.configbool('web', 'allowpull', True):
        log.debug('allow_pull failed')
        resp = exc.HTTPUnauthorized()
        resp.status = '%s %s' % (resp.code, 'pull not authorized for %s' % user)
        raise resp
    elif op == 'pull' or op is None: # op is None for interface requests
        return

    # enforce that you can only push using POST requests
    if req.environ['REQUEST_METHOD'] != 'POST':
        resp = exc.HTTPUnauthorized()
        resp.status = '%s %s' % (resp.code, 'push requires POST request')
        raise resp

    # require ssl by default for pushing, auth info cannot be sniffed
    # and replayed
    if c.configbool('web', 'push_ssl', True) and req.scheme != 'https':
        log.debug('puch_ssl fail')
        resp = exc.HTTPMethodNotAllowed()
        resp.status = '%s %s' % (resp.code, 'ssl required')
        raise resp

    if not user:
        resp = exc.HTTPUnauthorized()
        resp.status = '%s %s' % (resp.code, 'push not allowed')
        raise resp

    if 'allow_push' in permissions:
        append('allow_push')
    elif 'allow_push' not in permissions:
        deny = c.configlist('web', 'deny_push')
        if deny and (not user or deny == ['*'] or user in deny):
            log.debug('deny_push success')
            resp = exc.HTTPUnauthorized()
            resp.status = '%s %s' % (resp.code, 'push not allowed for user %s' % user)
            raise resp

        allow_push = c.configlist('web', 'allow_push')
        result = (not allow_push) or (allow_push == ['*'])
        if not (result or user in allow_push):
            log.debug('allow_push fail')
            resp = exc.HTTPUnauthorized()
            resp.status = '%s %s' % (resp.code, 'push not allowed for user %s' % user)
            raise resp
    else:
        allow_push = c.configlist('web', 'allow_push')
        if allow_push != ['*'] and user not in allow_read:
            append('allow_push')

def get_binary(suffix=None):
    binary = os.path.abspath(sys.argv[0])
    dirname = os.path.dirname(binary)
    if os.path.isfile(os.path.join(dirname, 'hgapp')):
        binary = os.path.join(dirname, 'hgapp')
    else:
        binary = binary.split('-')[0]
    if suffix:
        binary += '-%s' % suffix
    return binary

def cmd(*args, **kwargs):
    binary = kwargs.get('binary', get_binary())
    cwd = os.getcwd()
    os.chdir(kwargs.get('dirname', os.getcwd()))
    cmd = [binary]+list(args)
    if subprocess.call(cmd) != 0:
        print 'Error while running %s' % ' '.join(cmd)
    os.chdir(cwd)

def run(*args, **kwargs):
    cwd = os.getcwd()
    os.chdir(kwargs.get('dirname', os.getcwd()))
    if subprocess.call(list(args)) != 0:
        print 'Error while running %s' % ' '.join(cmd)
    os.chdir(cwd)

def hgrc(dirname):
    filename = os.path.join(dirname, '.hg', 'hgrc')
    return filename

def is_repository(dirname):
    for name in ('data', 'store'):
        exist = os.path.join(dirname, '.hg', name)
        if os.path.isdir(exist):
            return True
    return False

def edit_repo(dirname, section, **kwargs):
    cfg = ConfigObject()
    cfg.read(hgrc(dirname))
    for k, v in kwargs.items():
        cfg[section][k] = v
    with open(hgrc(dirname),'w') as f:
        cfg.write(f)

def init_repo(dest_dir, owner=None):
    log.warn('User %s creating repo at %s', owner, dest_dir)
    os.makedirs(dest_dir)
    cmd('hg', 'init', dest_dir)
    if owner:
        edit_repo(dest_dir, 'web', contact=owner)

def fork_repo(repo, dest_dir, owner, fork_path):
    if not os.path.isdir(dest_dir):
        log.warn('User %s cloning %s repo at %s', owner, fork_path, dest_dir)
        cmd('hg', 'clone', repo.root, dest_dir)
        edit_repo(dest_dir, 'paths', fork=fork_path)
        edit_repo(dest_dir, 'web', contact=owner,
                  description = 'Fork of %s' % fork_path)

def find_forks(path, user='*'):
    userdir = hgconfig['hg:config'].userdir
    forks = []
    filenames = glob.glob(os.path.join(userdir, user, '*', '.hg', 'hgrc'))
    for filename in filenames:
        cfg = ConfigObject()
        cfg.read(filename)
        if cfg.paths.fork and cfg.paths.default == path:
            filename = filename.split('/.hg/hgrc')[0]
            filename, name = os.path.split(filename)
            filename, user = os.path.split(filename)
            forks.append((user, name, cfg.web.description))
    return forks

def get_repo(path_info, user=None, cmd=''):
    path_info = path_info.strip('/')
    if '/' in path_info:
        name, path_info = path_info.split('/', 1)
    else:
        name = 'main'

    if name == 'config':
        return None, None, None

    repo = ui = c = None

    userdir = hgconfig['hg:config'].userdir
    if user and userdir:
        home = os.path.join(userdir, user)
        if not os.path.isdir(home):
            os.makedirs(home)
            cfg = ConfigObject()
            cfg.web = dict(allow_read='*', allow_push=user)
            fd = open(os.path.join(home, '.hgrc'), 'w')
            cfg.write(fd)
            fd.close()
        userdir = os.path.join(userdir, name)

    if userdir and os.path.isdir(userdir):
        c = hgconfig['hg:%s' % name]
        c.hgwebdir = userdir
        c.allow_push = name
        c.allow_read = name
    else:
        c = hgconfig['hg:%s' % name]
        c.name = name

    c.path_info = path_info
    if c.get('hgweb', None) and c.hgweb.endswith(path_info):
        repo = c.hgweb
    elif c.get('hgwebdir', None):
        path = os.path.join(c.hgwebdir, path_info)
        if is_repository(path):
            repo = path
        elif cmd.startswith('hg init'):
            repo = path

    ui = get_ui(c)

    return repo, ui, c

def get_ui(section):
    baseui = ui.ui()

    baseui.setconfig('ui', 'username', 'unknown')
    for k in ('report_untrusted', 'interactive'):
        baseui.setconfig('ui', k, section.get(k, 'off').as_bool('on', 'off'))

    baseui.setconfig('extensions', 'hgext.highlight', '')

    baseui.setconfig('web', 'style', 'gitweb')
    baseui.setconfig('web', 'encoding', 'utf-8')
    baseui.setconfig('web', 'push_ssl', section.push_ssl.as_bool('on', 'off'))
    baseui.setconfig('web', 'allowpull', section.get('allowpull', 'on').as_bool('on', 'off'))

    for k in ('allow_read', 'allow_push', 'deny_read', 'deny_push', 'require_groups', 'allow_archive'):
        v = section.get(k).as_list(sep=' ')
        if v:
            baseui.setconfig('web', k, ' '.join(v))

    baseui.setconfig('hooks', 'changegroup.hgapp_sphinx',
                          'python:hgapp.hooks.sphinx_update')

    filename = ''
    if section.hgweb:
        filename = isfile(hgrc(section.hgweb))
    elif section.hgwebdir:
        filename = isfile(section.hgwebdir, '.hgrc')
        sphinxroot = os.path.join(section.hgwebdir, '.sphinx')
        baseui.setconfig('hgapp', 'sphinxroot', sphinxroot)
        section.sphinx_docs = os.path.join(sphinxroot, 'html')

    if filename:
        cfg = config.config()
        cfg.read(filename)
        for k, v in cfg.items('web'):
            baseui.setconfig('web', k, v)

    return baseui

def isfile(*args):
     filename = os.path.join(*args)
     filename = os.path.realpath(filename)
     if os.path.isfile(filename):
         return filename
     return False

def load_config(*filenames):
    if not filenames:
        filenames = ['/etc/hgapp/hgapp.ini', os.path.expanduser('~/hgapp.ini'), 'hgapp.ini']
        filenames = [filename for filename in filenames if os.path.isfile(filename)]
        if not filenames:
            raise RuntimeError("Cannot find a hgapp.ini")
    filename = os.path.abspath(filenames[-1])
    cfg = ConfigObject(dict(here=os.path.dirname(filename),
                            __file__=filename))
    cfg.read([filename])
    cfg.__file__ = filename
    return cfg
